import click, os, subprocess, tempfile
from toolbox.logger import Log
from .aws_config import AWSconfig
from . import iam_nongc
from configstore.configstore import Config
from toolbox.misc import set_terminal_width

CONFIG = Config('awstools')
LATEST = CONFIG.get_profile('latest')

if LATEST is None:
    LATEST = 'default'
else:
    LATEST = LATEST['config']['role']

CONTEXT_SETTINGS = {'help_option_names':['-h','--help'], 'max_content_width': set_terminal_width(), 'ignore_unknown_options': True, 'allow_extra_args': True}

@click.command(help="run any awscli (aws) command while leveraging awstools profile functionality; append --help for usage on cli commands", context_settings=CONTEXT_SETTINGS)
@click.argument('awscli_command', nargs=-1, type=str, required=False)
@click.option('-r', '--region', 'aws_region', help='AWS region to authenticate against', required=False, default='us-east-1')
@click.pass_context
def cli(ctx, awscli_command, aws_region):
    aws_profile_name = ctx.obj['PROFILE']
    if aws_profile_name == 'default':
        check_env()
    AWS = get_AWScli(aws_profile_name, aws_region)
    CMD = ' '.join(awscli_command) # convert tuple to string
    RESULT = AWS.run_cmd(f"aws --region {aws_region} {CMD}")
    print(RESULT)

class AWScli():
    def __init__(self, aws_profile_name, in_boundary, aws_region='us-east-1'):
        self.INB = 'non-gc'
        self.AWS_PROFILE = aws_profile_name
        self.AWS_REGION = aws_region
        self.CONFIG = AWSconfig()
        self.CONFIG.unset_aws_profile()
        if self.CONFIG.profile_or_role(aws_profile_name):
            self.SESSION = iam_nongc._authenticate(aws_profile_name, aws_region)
        else:
            self.SESSION = iam_nongc._assume_role(aws_profile_name=aws_profile_name)[0]

    def get_session_creds(self):
        CREDS = self.SESSION.get_credentials()
        CREDS = CREDS.get_frozen_credentials()
        ACCESS_KEY = CREDS.access_key
        SECRET_KEY = CREDS.secret_key
        TOKEN = CREDS.token
        return ACCESS_KEY, SECRET_KEY, TOKEN
    
    def creds_to_env(self):
        KEY, SECRET, TOKEN = self.get_session_creds()
        os.environ['AWS_ACCESS_KEY_ID'] = KEY
        os.environ['AWS_SECRET_ACCESS_KEY'] = SECRET
        if TOKEN is not None:
            os.environ['AWS_SESSION_TOKEN'] = TOKEN
    
    def run_cmd(self, cmd):
        self.CONFIG.unset_aws_profile()
        self.creds_to_env()
        RESULT = run_command(cmd)
        return RESULT

def get_AWScli(aws_profile_name, aws_region='us-east-1'):
    inb = False
    return AWScli(aws_profile_name, inb, aws_region)

def run_command(command):
    TEMP = tempfile.NamedTemporaryFile(delete=False)
    TEMP.close()
    PATH = TEMP.name
    CMD = command + " 1> " + PATH + " 2>" + PATH
    PROC = subprocess.run(CMD, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
    with open(PATH, 'r') as f:
        DATA = f.read()
    os.unlink(PATH)
    return DATA
    
def check_env(): # ensure none of the AWS env variables is set
    COUNT = 0
    try:
        os.environ['AWS_SECRET_KEY_ID'] 
    except KeyError:
        COUNT = COUNT + 1
    try:
        os.environ['AWS_SECRET_ACCESS_KEY']
    except KeyError:
        COUNT = COUNT + 1
    try:
        os.environ['AWS_SESSION_TOKEN']
    except KeyError:
        COUNT = COUNT + 1
    if COUNT != 3:    
        Log.critical('Please unset AWS_SECRET_KEY_ID, AWS_SECRET_ACCESS_KEY and/or AWS_SESSION_TOKEN before proceeding')



from os import environ, unsetenv, mkdir, path
from pathlib import Path
from configparser import ConfigParser
from toolbox.logger import Log
from toolbox.getpass import get_secure_string

class AWSconfig():
    def __init__(self, aws_dir=f"{environ['HOME']}/.aws/", aws_config="config", aws_credentials="credentials"):
        self.AWS_DIR = aws_dir
        self.AWS_CONFIG_FILE = self.verify_path(f"{aws_dir}{aws_config}")
        self.AWS_CREDS_FILE = self.verify_path(f"{aws_dir}{aws_credentials}")

    def profile_or_role(self, profile_name):
        ROLE_ARN = self.get_from_config('creds', 'role_arn', 'string', "", profile_name)
        if ROLE_ARN == "" or ROLE_ARN is None:
            return True # no role_arn so the profile must be for the main account
        else:
            return False # role_arn fonud so the profile must be for a role

    def add_aws_profile(self, aws_access_key_id, aws_secret_access_key, aws_region, aws_output, profile_name):
        self.update_aws_config('creds', profile_name, 'aws_access_key_id', aws_access_key_id)
        self.update_aws_config('creds', profile_name, 'aws_secret_access_key', aws_secret_access_key)
        self.update_aws_config('config', profile_name, 'region', aws_region)
        self.update_aws_config('config', profile_name, 'output', aws_output)

    def add_aws_profile_saml(self, aws_access_key_id, aws_secret_access_key, aws_session_token, aws_region, aws_output, profile_name):
        self.update_aws_config('creds', profile_name, 'aws_access_key_id', aws_access_key_id)
        self.update_aws_config('creds', profile_name, 'aws_secret_access_key', aws_secret_access_key)
        self.update_aws_config('creds', profile_name, 'aws_session_token', aws_session_token)
        self.update_aws_config('config', profile_name, 'region', aws_region)
        self.update_aws_config('config', profile_name, 'output', aws_output)

    def config_add_role(self, profile_name, role_arn, creds_profile):
        self.update_aws_config('config', profile_name, 'role_arn', role_arn)
        self.update_aws_config('config', profile_name, 'source_profile', creds_profile)
    
    def add_sim_config(self, sim_region, token, iam_profile, profile_name):
        self.update_aws_config('config', profile_name, 'region', sim_region)
        self.update_aws_config('creds', profile_name, 'aws_session_token', token)
        self.update_aws_config('creds', profile_name, 'source_profile', iam_profile)

    def update_aws_config(self, config_type, profile_name, property_name, property_value):
        CONFIG, CONFIG_FILE = self.choose_configparser(config_type)
        if config_type == "config":
            profile_name = "profile " + profile_name
        try:
            CONFIG.add_section(profile_name)
        except:
            pass # this means the section already exists
        CONFIG.set(profile_name, property_name, property_value)
        with open(CONFIG_FILE, 'w') as CONFIG_FILE_STREAM:
            CONFIG.write(CONFIG_FILE_STREAM) 

    def get_from_config(self, config_type, config_property_name, alternative_method=None, alternative_method_value=None, profile_name=None, supplied_value=None):
        CONFIG, CONFIG_FILE = self.choose_configparser(config_type)
        if profile_name in CONFIG and config_property_name in CONFIG[profile_name]:
            RESULT = CONFIG[profile_name][config_property_name]
        else:
            RESULT = None
        if RESULT is None:
            if alternative_method == 'string':
                RESULT = alternative_method_value
            elif alternative_method == 'input':
                RESULT = input(alternative_method_value)
            elif alternative_method == 'secure':
                RESULT = get_secure_string(config_property_name, alternative_method_value)
            else:
                pass
        if supplied_value != None:
            if RESULT != supplied_value:
                RESULT = supplied_value
        return RESULT

    def choose_configparser(self, config_type):
        CONFIG = ConfigParser() 
        if config_type == "creds":
            CONFIG_FILE = self.AWS_CREDS_FILE
        elif config_type == "config":
            CONFIG_FILE = self.AWS_CONFIG_FILE
        else:
            Log.critical("invalid config type")
        CONFIG.read(CONFIG_FILE)
        return CONFIG, CONFIG_FILE

    def verify_path(self, path_to_file):
        try:
            PATH_TO_FOLDER = '/'.join(path_to_file.split('/')[:-1])
            if not path.exists(PATH_TO_FOLDER):
                mkdir(PATH_TO_FOLDER)
            Path(path_to_file).touch()
            return path_to_file
        except OSError:
                Log.critical("failed to create .aws directory")

    def unset_aws_profile(self, *args):
        if args is None:
            if 'AWS_ACCESS_KEY_ID' in os.environ:
                del environ['AWS_ACCESS_KEY_ID']
            if 'AWS_SECRET_ACCESS_KEY' in os.environ:
                del environ['AWS_SECRET_ACCESS_KEY']
            if 'AWS_DEFAULT_REGION' in os.environ:
                del environ['AWS_DEFAULT_REGION']
            if 'AWS_SESSION_TOKEN' in os.environ:
                del environ['AWS_SESSION_TOKEN']
            if 'AWS_OUTPUT' in os.environ:
                del environ['AWS_OUTPUT']
            if 'AWS_ROLE_ARN' in os.environ:
                del environ['AWS_ROLE_ARN']
            if 'AWS_SOURCE' in os.environ:
                del environ['AWS_SOURCE']
        else:
            for arg in args:
                unsetenv(arg)
        try:
            del environ['AWS_ACCESS_KEY_ID']
            del environ['AWS_SECRET_ACCESS_KEY']
            del environ['AWS_SESSION_TOKEN']
        except:
            pass

    def set_aws_profile(self, aws_session_token, aws_region, aws_output):
        environ['AWS_SESSION_TOKEN'] = aws_session_token
        environ['AWS_DEFAULT_REGION'] = aws_region
        environ['AWS_OUTPUT'] = aws_output

    def set_env_variables(self, aws_access_key_id, aws_secret_access_key, aws_session_token):
        environ['AWS_ACCESS_KEY_ID'] = aws_access_key_id
        environ['AWS_SECRET_ACCESS_KEY'] = aws_secret_access_key
        environ['AWS_SESSION_TOKEN'] = aws_session_token

    def display(self, type, profile_name=None):
        AWS_CONFIG, AWS_FILE = self.choose_configparser(type)
        if profile_name is None:
            print(f'\n-------------------- aws {type} for all profiles --------------------')
            for PROFILE in AWS_CONFIG:
                if PROFILE != 'DEFAULT': # aws uses lowercase default, remove redundant DEFAULT entry from output
                    print(f"    [{PROFILE}]")
                for PROPERTY in AWS_CONFIG[PROFILE]:
                    print(f"    {PROPERTY}={AWS_CONFIG[PROFILE][PROPERTY]}")
                print('')
            print('')
        else:
            if profile_name in AWS_CONFIG:
                print(f'\n-------------------- aws {type} for profile {profile_name} --------------------')
                print(f"    [{profile_name}]")
                for PROPERTY in AWS_CONFIG[profile_name]:
                        print(f"    {PROPERTY}={AWS_CONFIG[profile_name][PROPERTY]}")
                print('')
            else:
                return



from toolbox.logger import Log
from configstore.configstore import Config
from .aws_config import AWSconfig
from . import iam_nongc
import datetime
from tabulate import tabulate

CONFIGSTORE = Config('awstools')

class RDSclient():

    def __init__(self, aws_profile_name, aws_region='us-east-1', in_boundary=True, cache_only=False):
        self.CONFIG = AWSconfig()
        self.CONFIGSTORE = Config('awstools')
        self.CACHE_UPDATE_INTERVAL = 60*60*24*7 # update cache every week
        self.AWS_PROFILE = aws_profile_name
        self.AWS_REGION = aws_region
        self.INB = in_boundary
        self.CACHE_ONLY = cache_only
        if not self.CACHE_ONLY:
            self.get_session()
            
    def get_session(self):
        RES = 'non-gc'
        if RES == 'non-gc':
            if self.CONFIG.profile_or_role(self.AWS_PROFILE):
                self.SESSION = iam_nongc._authenticate(self.AWS_PROFILE, self.AWS_REGION)
            else:
                self.SESSION = iam_nongc._assume_role(aws_profile_name=self.AWS_PROFILE)[0]
        return self.SESSION

    def show_rds_instances(self, aws_profile_name):
        CACHE = self.get_cached_rds_instances(aws_profile_name)
        if CACHE is not None:
            return CACHE
        RDS_INSTANCES = self.get_rds_instances()
        if RDS_INSTANCES != {}:
            self.cache_rds_instances(RDS_INSTANCES, aws_profile_name)
        return RDS_INSTANCES

    def describe(self, database):
        self.get_session()
        CLIENT = self.SESSION.client('rds')
        RESPONSE = CLIENT.describe_db_instances(DBInstanceIdentifier=database)
        return RESPONSE

    def get_rds_instances(self):
        C=0
        self.get_session()
        CLIENT = self.SESSION.client('rds')
        RESPONSE = CLIENT.describe_db_instances()
        RDS_INSTANCES = {}
        for INSTANCE in RESPONSE['DBInstances']:
            C = C + 1
            RDS_DICT = {}
            RDS_DICT['DBInstanceArn'] = INSTANCE['DBInstanceArn']
            RDS_DICT['Engine'] = INSTANCE['Engine']
            RDS_DICT['DBInstanceStatus'] = INSTANCE['DBInstanceStatus']
            try:
                RDS_DICT['DBName'] = INSTANCE['DBName']
            except:
                RDS_DICT['DBName'] = "N/A"
            RDS_INSTANCES[INSTANCE['DBInstanceIdentifier']] = RDS_DICT
        if RDS_INSTANCES == {}:
            Log.warn('No RDS instances found in the target AWS account')
        return RDS_INSTANCES

    def get_cached_rds_instances(self, aws_profile_name):
        if aws_profile_name in self.CONFIGSTORE.PROFILES:
            if 'cached_rds_instances' in self.CONFIGSTORE.PROFILES[aws_profile_name]['metadata']:
                return self.CONFIGSTORE.PROFILES[aws_profile_name]['metadata']['cached_rds_instances'][self.AWS_REGION]
        else:
            return None

    def cache_rds_instances(self, instances, aws_profile_name):
        TIMESTAMP = str(datetime.datetime.now().timestamp())
        if aws_profile_name not in self.CONFIGSTORE.PROFILES:
            self.CONFIGSTORE.create_profile(aws_profile_name)
        if 'cached_rds_instances' not in self.CONFIGSTORE.PROFILES[aws_profile_name]['metadata']:
            self.CONFIGSTORE.PROFILES[aws_profile_name]['metadata']['cached_rds_instances'] = {}
        self.CONFIGSTORE.PROFILES[aws_profile_name]['metadata']['cached_rds_instances'] = instances
        self.CONFIGSTORE.PROFILES[aws_profile_name]['metadata']['cached_rds_instances']['last_cache_update'] = TIMESTAMP
        self.CONFIGSTORE.update_profile(self.CONFIGSTORE.PROFILES[aws_profile_name])

    def refresh(self, type, aws_profile_name):
        if type == 'cached_rds_instances':
            Log.info("caching rds instances...")
            RDS_INSTANCES = self.get_rds_instances()
            DICT = {}
            DICT[self.AWS_REGION] = RDS_INSTANCES
            self.cache_rds_instances(DICT, aws_profile_name)
        self.CONFIGSTORE = Config('awstools')

    def auto_refresh(self, aws_profile_name):
        TIME_NOW = datetime.datetime.now().timestamp()
        try:    
            RDS_INSTANCES_TS = self.CONFIGSTORE.PROFILES[aws_profile_name]['metadata']['cached_rds_instances']['last_cache_update']
        except KeyError:
            RDS_INSTANCES_TS = 0    
        if TIME_NOW - float(RDS_INSTANCES_TS) > self.CACHE_UPDATE_INTERVAL:
            Log.info('automatic refresh of rds instance cache initiated')
            self.refresh('cached_rds_instances', aws_profile_name)

    def get_region_from_profile(self, aws_profile_name):
        AWS_REGION = self.CONFIG.get_from_config('creds', 'region', profile_name=aws_profile_name)
        if AWS_REGION is None: # this is for when the user wants to use a profile which sources another profile for IAM creds
            CREDS_PROFILE = self.CONFIG.get_from_config('creds', 'source_profile', profile_name=aws_profile_name)
            AWS_REGION = self.CONFIG.get_from_config('config', 'region', profile_name=CREDS_PROFILE)
        if AWS_REGION is None:
            Log.critical("unable to get the correct region from the config")
        return AWS_REGION



import click
from  configstore.configstore_ctrl import _show as configshow
from .ec2 import _show as ec2show, orphaned_ebs_volumes as orphans, orphaned_ebs_snap_report as snap_report
from .s3 import _show as s3show, get_S3client
from .aws_config import AWSconfig
from .iam import get_latest_profile

@click.command(help='display configuration data for awstools and awscli')
@click.argument('target', required=True, type=click.Choice(['all_config', 'buckets', 'instances', 'public_ips', 'regions', 'all_cached', 'aws_config', 'aws_credentials', 'orphaned_ebs']))
@click.pass_context
def show(ctx, target):
    aws_profile_name = ctx.obj['PROFILE']
    aws_region_name = get_region(ctx, aws_profile_name)
    if aws_profile_name is None:
        aws_profile_name = 'default'
    if target == 'buckets':
        s3show('buckets', aws_profile_name, aws_region_name)
    if target == 'instances':
        ec2show('instances', aws_profile_name, aws_region_name, display=True)
    if target == 'public_ips':
        ec2show('public_ips', aws_profile_name, aws_region_name, display=True)
    if target == 'regions':
        ec2show('regions', aws_profile_name, aws_region_name, display=True)
    if target == 'all_cached':
        ec2show('all', aws_profile_name, aws_region_name, display=True)
    if target == 'all_config':
        configshow('awstools', aws_profile_name)
    if target == 'orphaned_ebs' or target == 'all':
        orphans(aws_profile_name)
        snap_report(aws_profile_name)
    if target == 'aws_config':
        AWS_CONFIG = AWSconfig()
        AWS_CONFIG.display('config', aws_profile_name)
    if target == 'aws_credentials':
        AWS_CONFIG = AWSconfig()
        AWS_CONFIG.display('creds', aws_profile_name)

def get_region(ctx, aws_profile_name):
    AWS_REGION = ctx.obj['REGION']
    if not AWS_REGION:
        S3 = get_S3client(aws_profile_name, 'us-east-1')
        AWS_REGION = S3.get_region_from_profile(aws_profile_name)
    return AWS_REGION




#!/usr/bin/env python3
import sys, click
from toolbox.logger import Log
from .s3 import s3
from .iam import iam
from .ec2 import ec2
from .rds import rds
from .awscli_wrapper import cli
from .awstools_show import show
from toolbox import misc
from toolbox.click_complete import complete_aws_regions, complete_aws_profiles
from toolbox.misc import debug
from .iam import get_latest_profile, get_latest_region
from .iam_nongc import update_latest_profile, force_cache, cache_all_hack
from .aws_config import AWSconfig
from .extract import extract

MESSAGE="AWS CLI Client" + misc.MOVE + "Current Profile: " + misc.GREEN + misc.UNDERLINE + get_latest_profile().upper() + misc.RESET + " Region: " + misc.GREEN + misc.UNDERLINE + get_latest_region(get_latest_profile()).upper() + misc.RESET

@click.group('aws', help=MESSAGE, context_settings={'help_option_names':['-h','--help'], 'max_content_width': misc.set_terminal_width()}, invoke_without_command=True)
@click.option('-p', '--profile', 'aws_profile_name', help='profile name to use when working with awstools', required=False, default=get_latest_profile(), shell_complete=complete_aws_profiles)
@click.option('-r', '--region', 'aws_region_name', help='region name to use when working with awstools', required=False, type=str, shell_complete=complete_aws_regions, default=None)
@click.option('-c', '--cache', 'cache', help="refresh all cache for region and tenant", is_flag=True, required=False, default=False, show_default=True)
@click.pass_context
def CLI(ctx, aws_profile_name, aws_region_name, cache):
    ctx.ensure_object(dict)
    ctx.obj['PROFILE'] = aws_profile_name
    ctx.obj['REGION'] = aws_region_name
    if cache is True:
        force_cache(aws_profile_name, aws_region_name)
        sys.exit()
    if aws_region_name != get_latest_region(aws_profile_name) and aws_region_name != None:
        CONFIG = AWSconfig()
        CONFIG.update_aws_config('creds', aws_profile_name, 'region', aws_region_name)
        CONFIG.update_aws_config('config', aws_profile_name, 'region', aws_region_name)
        cache_all_hack(aws_profile_name, aws_region_name)
    if aws_profile_name != get_latest_profile():
        update_latest_profile(aws_profile_name)
        cache_all_hack(aws_profile_name, aws_region_name)
    Log.setup('awstools', int(debug))
    pass
    
CLI.add_command(iam)
CLI.add_command(s3)
CLI.add_command(ec2)
CLI.add_command(rds)
CLI.add_command(cli)
CLI.add_command(show)
CLI.add_command(extract)

if __name__ == "__main__":
    cli(ctx)



import click, os, csv
from toolbox.logger import Log
from .ec2client import EC2client
from toolbox.menumaker import Menu
from tabulate import tabulate
from configstore.configstore import Config
from toolbox.misc import set_terminal_width
from .iam import get_latest_profile, get_latest_region

CONFIG = Config('awstools')

# just something I've always included to make sure I'm using the right input method based on python version. Not really needed anymore, but old habits die hard
try: input = raw_input
except NameError: raw_input = input

@click.group('ec2', invoke_without_command=True, help='ec2 to manage the instances and related configuration', context_settings={'help_option_names':['-h','--help'], 'max_content_width': set_terminal_width()})
@click.option('-m', '--menu', help='use the menu to perform EC2 actions', is_flag=True, show_default=True, default=False, required=False)
@click.pass_context
def ec2(ctx, menu):
    if menu:
        ctx.obj['MENU'] = True
    else:
        ctx.obj['MENU'] = False

@ec2.command(help='manually refresh ec2 cached data', context_settings={'help_option_names':['-h','--help']})
@click.pass_context
@click.argument('cache_type', required=False, default='all', type=click.Choice(['all', 'instances', 'public_ips', 'regions', 'auto']))
def refresh(ctx, cache_type):
    aws_profile_name = ctx.obj['PROFILE']
    aws_region = get_latest_region(aws_profile_name)
    _refresh(cache_type, aws_profile_name, aws_region)

def _refresh(cache_type, aws_profile_name, aws_region):
    try:    
        EC2 = get_EC2client(aws_profile_name, aws_region, auto_refresh=False)
        if cache_type == 'all':
            EC2.refresh(aws_profile_name, aws_region)
        if cache_type == 'instances':
            EC2.cache_instances(aws_profile_name, aws_region)
        if cache_type == 'public_ips':
            EC2.cache_public_ips(aws_profile_name, aws_region)
        if cache_type == 'regions':
            EC2.cache_regions(aws_profile_name, aws_region)
        if cache_type == 'auto':
            EC2.auto_refresh(aws_profile_name, aws_region)
        return True
    except:
        return False

@ec2.command(help='show the data stored in ec2 cache (instacnes, IPs, regions)', context_settings={'help_option_names':['-h','--help']})
@click.pass_context
@click.argument('target', required=False, default='all', type=click.Choice(['all', 'instances', 'public_ips', 'regions', 'orphaned_ebs', 'orphaned_snaps']))
def show(ctx, target):
    aws_profile_name = ctx.obj['PROFILE']
    aws_region = get_latest_region(aws_profile_name)
    if ctx.obj["MENU"] and target == "instances":
        CACHED_INSTANCES = {}
        DATA = []
        for PROFILE in CONFIG.PROFILES:
            if PROFILE == 'latest':
                continue
            if PROFILE == aws_profile_name:
                EC2 = get_EC2client(aws_profile_name, aws_region, auto_refresh=False, cache_only=True)
                RES = EC2.show_cache(aws_profile_name, 'cached_instances', aws_region, display=False)
                for data in RES:
                    NAME = data["PrivateDnsName"].ljust(50)
                    IP = data["PrivateIpAddress"]
                    ID = data["InstanceId"]
                    details = NAME + "\t" + IP + "\t" + ID
                    DATA.append(details)
                break
        INPUT = 'Login Manager'
        CHOICE = runMenu(DATA, INPUT)
        if CHOICE:
            IP = CHOICE[0].split("\t")[1]
            Log.info(f"you chose to SSH to {IP} which will be implemented later...")
    elif ctx.obj["MENU"] and target == "orphaned_ebs":
        INPUT = 'ORPHANED EBS VOLUMES'
        EC2 = EC2client(aws_profile_name, aws_region='us-east-1', in_boundary=False, cache_only=False)
        CLIENT = EC2.SESSION.client('ec2')
        VOLUMES = CLIENT.describe_volumes()
        DATA = []
        for VOLUME in VOLUMES['Volumes']:
            if VOLUME['State'] == 'available':
                DETAILS = VOLUME['VolumeId'] + "\t" + str(VOLUME['Size']) + "\t" + VOLUME['VolumeType'] + "\t" + VOLUME['State']
                DATA.append(DETAILS)
        CHOICE = runMenu(DATA, INPUT)
        if CHOICE:
            ID = CHOICE[0].split("\t")[0]
            Log.info(f"you chose to work with {ID} which will be implemented later...")
    else:
        _show(target, aws_profile_name, aws_region, display=True)

def _show(target, aws_profile_name, aws_region, display):
    try:
        EC2 = get_EC2client(aws_profile_name, aws_region, auto_refresh=False, cache_only=True)
        if target == 'instances' or target == 'all':
            EC2.show_cache(aws_profile_name, 'cached_instances', aws_region, display)
        if target == 'public_ips' or target == 'all':
            EC2.show_cache(aws_profile_name, 'cached_public_ips', aws_region, display)
        if target == 'regions' or target == 'all':
            EC2.show_cache(aws_profile_name, 'cached_regions', aws_region, display)
        if target == 'orphaned_ebs' or target == 'all':
            orphaned_ebs_volumes(aws_profile_name)
        if target == 'orphaned_snaps':
            orphaned_ebs_snap_report(aws_profile_name, delete=False)
        return True
    except:
        return False

def orphaned_ebs_volumes(aws_profile_name):
    EC2 = EC2client(aws_profile_name, aws_region='us-east-1', in_boundary=False, cache_only=False)
    CLIENT = EC2.SESSION.client('ec2')
    VOLUMES = CLIENT.describe_volumes()
    ORPHANED_VOLUMES = {'VolumeId': [], 'VolumeSize': [], 'VolumeType': [], 'State': [], 'CreateTime': []}
    for VOLUME in VOLUMES['Volumes']:
        if VOLUME['State'] == 'available':
            ORPHANED_VOLUMES['VolumeId'].append(VOLUME['VolumeId'])
            ORPHANED_VOLUMES['VolumeSize'].append(VOLUME['Size'])
            ORPHANED_VOLUMES['VolumeType'].append(VOLUME['VolumeType'])
            ORPHANED_VOLUMES['State'].append(VOLUME['State'])
            ORPHANED_VOLUMES['CreateTime'].append(VOLUME['CreateTime'])
    Log.info(f'orphaned_ebs_volumes count: {len(ORPHANED_VOLUMES["VolumeId"])}')
    Log.info(f"orphaned_ebs_volumes \n{tabulate(ORPHANED_VOLUMES, headers='keys', tablefmt='rst')}\n")
    return ORPHANED_VOLUMES

@ec2.group('delete', help="Delete EC2 objects such as instances, EBS volumes or snapshots", context_settings={'help_option_names':['-h','--help']})
def delete():
    pass

@delete.command('orphaned_ebs', help="delete orphaned EBS volumes or snapshots", context_settings={'help_option_names':['-h','--help']})
@click.option('-v', '--volume', help='flag to delete orphaned EBS volumes', required=False, is_flag=True)
@click.option('-s', '--snapshot', help='flag to delete orphaned EBS snapshots', required=False, is_flag=True)
@click.pass_context
def delete_orphaned_ebs_volumes(ctx, volume, snapshot):
    aws_profile_name = ctx.obj['PROFILE']
    EC2 = EC2client(aws_profile_name, aws_region='us-east-1', in_boundary=False, cache_only=False)
    CLIENT = EC2.SESSION.client('ec2')
    VOLUMES = CLIENT.describe_volumes()
    if ctx.obj["MENU"] and volume is True:
        INPUT = 'ORPHANED EBS VOLUMES'
        DATA = []
        for VOLUME in VOLUMES['Volumes']:
            if VOLUME['State'] == 'available':
                DETAILS = VOLUME['VolumeId'] + "\t" + str(VOLUME['Size']) + "\t" + VOLUME['VolumeType'] + "\t" + VOLUME['State']
                DATA.append(DETAILS)
        CHOICE = runMenu(DATA, INPUT)
        if CHOICE:
            ID = CHOICE[0].split("\t")[0]
            USER_INPUT = input(f"Are you sure you want to delete the unorphaned EBS volume {ID}? [y/n]: ")
            if USER_INPUT.lower() == "y":
                CLIENT.delete_volume(VolumeId=ID)
                Log.info(f"The unorphaned EBS volume {ID} has been deleted.")
            elif USER_INPUT.lower() == "n":
                Log.info(f"The unorphaned EBS volumes {ID} will not be deleted.")
            else:
                Log.info("Please enter a valid value (y/n).")
    elif volume is True:
        USER_INPUT = input("Are you sure you want to delete the unorphaned EBS volumes? [y/n]: ")
        if USER_INPUT.lower() == "y":
            ORPHANED_VOLUMES = orphaned_ebs_volumes()
            for VOLUME in ORPHANED_VOLUMES:
                CLIENT.delete_volume(VolumeId=VOLUME)
            Log.info("The unorphaned EBS volumes have been deleted.")
        elif USER_INPUT.lower() == "n":
            Log.info("The unorphaned EBS volumes will not be deleted.")
        else:
            Log.info("Please enter a valid value (y/n).")
    if snapshot is True:
        orphaned_ebs_snap_report(aws_profile_name, delete=True)

def get_EC2client(aws_profile_name, aws_region='us-east-1', auto_refresh=True, cache_only=False):
    CLIENT = EC2client(aws_profile_name, aws_region, False, cache_only)
    if auto_refresh:
        CLIENT.auto_refresh(aws_profile_name)
    return CLIENT

def runMenu(DATA, INPUT):
    COUNT = 0
    FINAL = []
    TITLE = f'EC2 Menu: {INPUT}'
    for data in DATA:
        COUNT = COUNT + 1
        RESULTS = []
        RESULTS.append(data)
        FINAL.append(RESULTS)
    SUBTITLE = f'showing {COUNT} available object(s)'
    JOINER = '\t\t'
    FINAL_MENU = Menu(FINAL, TITLE, JOINER, SUBTITLE)
    CHOICE = FINAL_MENU.display()
    return CHOICE

def orphaned_ebs_snap_report(aws_profile_name, delete=False):

    import boto3
    from botocore.exceptions import ClientError
    from datetime import datetime, timedelta, timezone
    import sys

    CSV = '/tmp/' + aws_profile_name + '_snapshots.csv'
    IGNORE_DAYS = 0
    MYDATA = []
    EC2 = EC2client(aws_profile_name, aws_region='us-east-1', in_boundary=False, cache_only=False)
    EC2 = EC2.SESSION.client('ec2')
    SNAPSHOTS = EC2.describe_snapshots(OwnerIds=['self']).get('Snapshots', [])
    SNAPSHOTS = []
    RESPONSE = EC2.describe_snapshots(OwnerIds=['self'], MaxResults=1000)
    while True:
        SNAPSHOTS.extend(RESPONSE.get('Snapshots', []))
        TOKEN = RESPONSE.get('NextToken', '')
        if TOKEN == '':
            break
        RESPONSE = EC2.describe_snapshots(NextToken=TOKEN, OwnerIds=['self'], MaxResults=1000)
    if SNAPSHOTS:
        SNAPSHOTS.sort(key=lambda snapshot: snapshot['VolumeId'] + str(snapshot['StartTime']))
        V_PREV = None
        NUM_VOLUMES = 0
        NUM_VOLUMES_IGNORED = 0
        NUM_SNAP_VOL_IGNORED = 0
        NUM_SNAP_TOO_NEW = 0
        NUM_SNAP_COMPLIANCE = 0
        NUM_SNAP_AMI = 0
        NUM_SNAP_ORPHANS = 0
        IGNORE_VOLUME = False
        DELETE_CONFIRMED = False
        Log.info(f"please wait while generating a report; this might take a minute depending on the number of snapshots for {aws_profile_name}")
        for row in SNAPSHOTS:
            V = row['VolumeId']
            SID = row['SnapshotId']
            V_SIZE = row['VolumeSize']
            if V != V_PREV or V == 'vol-ffffffff':
                NUM_VOLUMES += 1
                V_PREV = V
                IGNORE_VOLUME = False
                if V != 'vol-ffffffff':
                    try:
                        VOLUMES = EC2.describe_volumes(VolumeIds=[V]).get('Volumes', [])
                        if VOLUMES:
                            NUM_VOLUMES_IGNORED += 1
                            IGNORE_VOLUME = True
                    except ClientError as e:
                        CODE = e.response.get('Error', {}).get('Code')
                        if CODE != 'InvalidVolume.NotFound':
                            raise
            if IGNORE_VOLUME:
                NUM_SNAP_VOL_IGNORED += 1
            else:
                TIMESTAMP = row['StartTime']
                if TIMESTAMP > datetime.now(timezone.utc) - timedelta(days=IGNORE_DAYS):
                    NUM_SNAP_TOO_NEW += 1
                    continue
                for tag in row.get('Tags', []):
                    if (tag.get('Key') == 'aws:backup:source-resource' or
                        tag.get('Key') == 'dlm:managed'):
                        NUM_SNAP_COMPLIANCE += 1
                        break
                else:
                    IMAGES = EC2.describe_images(
                                 Filters=[{'Name':'block-device-mapping.snapshot-id', 'Values':[SID]}]
                             ).get('Images', [])
                    if IMAGES:
                        NUM_SNAP_AMI += 1
                        continue  # Done with this snapshot as it's attached to an AMI, go to next row
                    NUM_SNAP_ORPHANS += 1
                    TS_STR = str(TIMESTAMP).split('.')[0].split('+')[0]  # Strip off ms & timezone info
                    if delete is True:
                        try:
                            if not DELETE_CONFIRMED:
                                print("You have chosen to delete matching snapshots.  It's recommended that you first run this "
                                      "script without the 'delete' function to get a report of snapshots that will be deleted.\n"
                                )
                                RESPONSE = raw_input('Are you sure you want to continue [Y/N]?: ').strip()
                                if RESPONSE == 'Y':
                                    DELETE_CONFIRMED = True
                                else:
                                    sys.exit()
                            EC2.delete_snapshot(SnapshotId=SID)
                        except ClientError as e:
                            if e.response.get('Error', {}).get('Code') != 'DryRunOperation':
                                raise  # Re-raise other exceptions
                        print('Deleted => ', end='')
                        Log.info(f'({V} / {SID}) => {TS_STR}')
                    MYDICT = {}
                    DESC = row.get('Description', '')
                    for tag in row.get('Tags', []):
                        if tag.get('Key') == 'Name':
                            DESC = tag.get('Value', '')
                            break
                    DESC = (DESC[:75] + '..') if len(DESC) > 75 else DESC
                    MYDICT['VolumeId'] = V
                    MYDICT['SnapshotId'] = SID
                    MYDICT['Description'] = DESC
                    MYDICT['Timestamp'] = TS_STR
                    MYDICT['VolumeSize'] = V_SIZE
                    MYDATA.append(MYDICT)
        print()
        if MYDATA:
            Log.info(f"\n{tabulate(MYDATA, headers='keys', tablefmt='fancy_grid')}")
            save_results(MYDATA, CSV)
            print(f"\nCSV RESULTS: {CSV}")
        print(f'\nTOTAL VOLUMES: {NUM_VOLUMES}')
        print(f'\tIgnored because volume still exists: {NUM_VOLUMES_IGNORED}')
        print(f'\nTOTAL SNAPSHOTS: {len(SNAPSHOTS)}')
        print(f'\tIgnored because volume was ignored: {NUM_SNAP_VOL_IGNORED}')
        print(f'\tOf remaining, ignored because too new: {NUM_SNAP_TOO_NEW}')
        print(f'\tOf remaining, ignored because managed by AWS Backup or Amazon Data Lifecycle Manager: {NUM_SNAP_COMPLIANCE}')
        print(f'\tOf remaining, ignored because linked to an AMI: {NUM_SNAP_AMI}')
        print(f"\tRemaining Orphans reported: {NUM_SNAP_ORPHANS}")
        print()
    else:
        Log.info(f'No snapshots in this region owned by {aws_profile_name}.')

def save_results(data, csvfile):
    ROWS = ['VolumeId', 'SnapshotId', 'Description', 'Timestamp', 'VolumeSize']
    with open(csvfile, 'w') as CSV:
        writer = csv.DictWriter(CSV, fieldnames=ROWS)
        writer.writeheader()
        writer.writerows(data)




from toolbox.logger import Log
from configstore.configstore import Config
from .aws_config import AWSconfig
from . import iam_nongc
import datetime
from tabulate import tabulate

CONFIGSTORE = Config('awstools')

class EC2client():

    def __init__(self, aws_profile_name, aws_region='us-east-1', in_boundary=True, cache_only=False):
        self.CONFIG = AWSconfig()
        self.CONFIGSTORE = Config('awstools')
        self.CACHE_UPDATE_INTERVAL = 60*60*24*7 # update cache every week
        self.AWS_PROFILE = aws_profile_name
        self.AWS_REGION = aws_region
        self.INB = in_boundary
        self.CACHE_ONLY = cache_only
        if not self.CACHE_ONLY:
            self.get_session()

    def get_session(self):
        Log.info(f"connecting to EC2 as {self.AWS_PROFILE} via {self.AWS_REGION}...")
        RES = 'non-gc'
        if RES == 'non-gc':
            if self.CONFIG.profile_or_role(self.AWS_PROFILE):
                self.SESSION = iam_nongc._authenticate(self.AWS_PROFILE, self.AWS_REGION)
            else:
                self.SESSION = iam_nongc._assume_role(aws_profile_name=self.AWS_PROFILE)[0]
        return self.SESSION

    def auto_refresh(self, aws_profile_name='default'):
        TIME_NOW = datetime.datetime.now().timestamp()
        try:
            INSTANCES_TS = self.CONFIGSTORE.PROFILES[aws_profile_name]['metadata']['cached_instances']['last_cache_update']
        except KeyError:
            INSTANCES_TS = 0
        if TIME_NOW - float(INSTANCES_TS) > self.CACHE_UPDATE_INTERVAL:
            self.cache_instances(aws_profile_name)
        try: 
            IPS_TS = self.CONFIGSTORE.PROFILES[aws_profile_name]['metadata']['cached_public_ips']['last_cache_update']
        except KeyError:
            IPS_TS = 0
        if TIME_NOW - float(IPS_TS) > self.CACHE_UPDATE_INTERVAL:
            self.cache_public_ips(aws_profile_name)
        try:
            REGIONS_TS = self.CONFIGSTORE.PROFILES[aws_profile_name]['metadata']['cached_regions']['last_cache_update']
        except KeyError:
            REGIONS_TS = 0
        if TIME_NOW - float(REGIONS_TS) > self.CACHE_UPDATE_INTERVAL:
            self.cache_regions(aws_profile_name)
        self.CONFIGSTORE = Config('awstools')

    def refresh(self, aws_profile_name='default'):
        self.cache_instances(aws_profile_name)
        self.cache_public_ips(aws_profile_name)
        self.cache_regions(aws_profile_name)

    def cache_instances(self, aws_profile_name='default'):
        TIMESTAMP = str(datetime.datetime.now().timestamp())
        if aws_profile_name not in self.CONFIGSTORE.PROFILES:
            self.CONFIGSTORE.create_profile(aws_profile_name)
        if self.CACHE_ONLY and hasattr(self, 'SESSION') is False:
            self.get_session()
        CLIENT = self.SESSION.client('ec2')
        Log.info("caching ec2 instances...")
        INSTANCES = self.get_instances(CLIENT)
        INSTANCES_CACHE = {}
        for INSTANCE in INSTANCES:
            INSTANCE_NAME = None
            for INDEX in range(len(INSTANCE['Tags'])):
                if INSTANCE['Tags'][INDEX]['Key'] == 'Name':
                    INSTANCE_NAME = INSTANCE['Tags'][INDEX]['Value']
            if INSTANCE_NAME is None:
                INSTANCE_NAME = INSTANCE['InstanceId']
            INSTANCE_DICT = {
                'InstanceName': INSTANCE_NAME,
                'ImageId': INSTANCE['ImageId'],
                'InstanceId': INSTANCE['InstanceId'],
                'InstanceType': INSTANCE['InstanceType'],
                'PrivateDnsName': INSTANCE['PrivateDnsName'],
                'PrivateIpAddress': INSTANCE['PrivateIpAddress']
            }
            INSTANCES_CACHE[INSTANCE_NAME] = INSTANCE_DICT
        INSTANCES_CACHE['last_cache_update'] = TIMESTAMP
        DICT = {}
        DICT[self.AWS_REGION] = INSTANCES_CACHE 
        self.CONFIGSTORE.update_metadata(DICT, 'cached_instances', aws_profile_name, False)
        self.CONFIGSTORE = Config('awstools')
    
    def get_instances(self, EC2):
        RESULT = EC2.describe_instances(
            Filters = [
                {
                    'Name': 'instance-state-name',
                    'Values': ['running'] 
                },
                {
                    'Name': 'tag:Name',
                    'Values': ['*']
                }
            ]
        )
        INSTANCES = []
        RESERVATIONS = RESULT['Reservations']
        for INDEX in range(len(RESERVATIONS)):
            RESERVATION = RESERVATIONS[INDEX]
            INSTANCE_LIST = RESERVATION['Instances']
            for INSTANCE in INSTANCE_LIST:
                INSTANCES.append(INSTANCE)
        return INSTANCES

    def cache_public_ips(self, aws_profile_name='default'):
        TIMESTAMP = str(datetime.datetime.now().timestamp())
        if aws_profile_name not in self.CONFIGSTORE.PROFILES:
            self.CONFIGSTORE.create_profile(aws_profile_name)
        if self.CACHE_ONLY and hasattr(self, 'SESSION') is False:
            self.get_session()
        CLIENT = self.SESSION.client('ec2')
        Log.info("caching ec2 public ips...")
        IPS = CLIENT.describe_addresses()
        IPS_CACHE = {}
        for IP in IPS['Addresses']:
            IP_ADDRESS = IP['PublicIp']
            if 'PrivateIpAddress' in IP:
                IP_PRIVATE = IP['PrivateIpAddress'],
            else:
                IP_PRIVATE = ""
            IP_DICT = {
                'Domain': IP['Domain'],
                'PrivateIpAddress': IP_PRIVATE,
                'NetworkBorderGroup': IP['NetworkBorderGroup']
            }
            IPS_CACHE[IP_ADDRESS] = IP_DICT
        IPS_CACHE['last_cache_update'] = TIMESTAMP
        self.CONFIGSTORE.update_metadata(IPS_CACHE, 'cached_public_ips', aws_profile_name, True)
        self.CONFIGSTORE = Config('awstools')

    def cache_regions(self, aws_profile_name='default'):
        TIMESTAMP = str(datetime.datetime.now().timestamp())
        if aws_profile_name not in self.CONFIGSTORE.PROFILES:
            self.CONFIGSTORE.create_profile(aws_profile_name)
        if self.CACHE_ONLY and hasattr(self, 'SESSION') is False:
            self.get_session()
        CLIENT = self.SESSION.client('ec2')
        Log.info("caching ec2 regions...")
        REGIONS = CLIENT.describe_regions()
        REGIONS_CACHE = {}
        for REGION in REGIONS['Regions']:
            REGION_NAME = REGION['RegionName']
            REGION_DICT = {
                'Endpoint': REGION['Endpoint'],
                'OptInStatus': REGION['OptInStatus']
            }
            REGIONS_CACHE[REGION_NAME] = REGION_DICT
        REGIONS_CACHE['last_cache_update'] = TIMESTAMP
        self.CONFIGSTORE.update_metadata(REGIONS_CACHE, 'cached_regions', aws_profile_name, True)
        self.CONFIGSTORE = Config('awstools')

    def show_cache(self, aws_profile_name, type, aws_region, display):
        self.auto_refresh(aws_profile_name)
        DATA = []
        if type == 'cached_instances':
            for ENTRY in self.CONFIGSTORE.PROFILES[aws_profile_name]['metadata'][type][aws_region]:
                DATA_ENTRY = {}
                if ENTRY != 'last_cache_update':
                    DATA_ENTRY['Name/ID'] = ENTRY
                    for PROPERTY in self.CONFIGSTORE.PROFILES[aws_profile_name]['metadata'][type][aws_region][ENTRY]:
                        DATA_ENTRY[PROPERTY] = self.CONFIGSTORE.PROFILES[aws_profile_name]['metadata'][type][aws_region][ENTRY][PROPERTY]
                    DATA.append(DATA_ENTRY)
            if DATA != []:
                if display is True:
                    Log.info(f"{type}\n{tabulate(DATA, headers='keys', tablefmt='rst')}\n")
                return DATA
        else:
            for ENTRY in self.CONFIGSTORE.PROFILES[aws_profile_name]['metadata'][type]:
                DATA_ENTRY = {}
                if ENTRY != 'last_cache_update':
                    DATA_ENTRY['Name/ID'] = ENTRY
                    for PROPERTY in self.CONFIGSTORE.PROFILES[aws_profile_name]['metadata'][type][ENTRY]:
                        DATA_ENTRY[PROPERTY] = self.CONFIGSTORE.PROFILES[aws_profile_name]['metadata'][type][ENTRY][PROPERTY]
                    DATA.append(DATA_ENTRY)
            if DATA != []:
                if display is True:
                    Log.info(f"{type}\n{tabulate(DATA, headers='keys', tablefmt='rst')}\n")
                return DATA

    def show_all_cache(self, aws_profile_name, aws_region):
        self.show_cache(aws_profile_name, 'cached_instances', aws_region, display=False)
        self.show_cache(aws_profile_name, 'cached_public_ips', aws_region, display=False)
        self.show_cache(aws_profile_name, 'cached_regions', aws_region, display=False)



import sys, os

try:
    import importlib_resources as resources
except:
    from importlib import resources

def get_accounts(target_names):

    TARGETS = []
    RESULT = 'non-gc'
    MY_RESOURCES = resources.files("toolbox")
    if 'prod' in RESULT:
        DATA = (MY_RESOURCES / "prod_accounts.lst")
    else:
        DATA = (MY_RESOURCES / "stage_accounts.lst")

    target_names = ''.join(target_names)
    with open(DATA) as f:
        CONTENTS = f.read()
    for line in CONTENTS.split('\n'):
        if line:
            item = line.split(",")[0]
            value = line.split(",")[1]
            if item == target_names:
                TARGETS.append(line.split(",")[1])
    if TARGETS:
        return TARGETS
    else:
        return target_names




import click, boto3, string, random, subprocess
from .aws_config import AWSconfig
from .accounts import get_accounts
from toolbox.logger import Log
from os import environ, chmod, stat
from configstore.configstore import Config
from toolbox.misc import set_terminal_width

CONFIG = AWSconfig()

@click.command(help='login using existing IAM creds or add new creds to config', context_settings={'help_option_names':['-h','--help'], 'max_content_width': set_terminal_width()})
@click.option('-r', '--region', 'aws_region', help='aws region to connect to', required=False, default='us-east-1')
@click.option('-o', '--output', 'aws_output', help='output type for awscli', required=False, default='json')
@click.pass_context
def authenticate(ctx, aws_region, aws_output):
    profile_name = ctx.obj['PROFILE']
    RESULT = _authenticate(profile_name, aws_region, aws_output)
    if RESULT is None:
        MSG = 'setup failed. please verify supplied credentials and try again. settings were not saved'
        LINK = 'https://github.com/stacksc/goat'
        CMD = None
        SUBTITLE = 'CRITICAL'
        TITLE = 'GOAT'
        Log.notify(MSG, TITLE, SUBTITLE, LINK, CMD)
        Log.critical(MSG)
    else:
        Log.info("credentials saved successfully")
        cache_all_hack(profile_name, aws_region)
        Log.info(f"you can now use your new profile with 'aws --profile {profile_name}")
        update_latest_profile(profile_name)
        MSG = f'{profile_name} credentials saved successfully!'
        LINK = None
        CMD = None
        SUBTITLE = 'INFO'
        TITLE = 'GOAT'
        Log.notify(MSG, TITLE, SUBTITLE, LINK, CMD)

# worker function to make the method portable
def _authenticate(profile_name, aws_region='us-east-1', aws_output='json'):
    CONFIG.unset_aws_profile()
    AWS_REGION = CONFIG.get_from_config('config', 'region', 'string', 'us-east-1', profile_name, aws_region)
    AWS_OUTPUT = CONFIG.get_from_config('config', 'output', 'string', 'json', profile_name, aws_output)
    AWS_KEY_ID, AWS_SECRET_ACCESS_KEY = get_env_variables()
    if AWS_KEY_ID is None or AWS_SECRET_ACCESS_KEY is None:
        AWS_KEY_ID = CONFIG.get_from_config('creds', 'aws_access_key_id', 'input', 'Please enter AWS_KEY_ID: ', profile_name)
        AWS_SECRET_ACCESS_KEY = CONFIG.get_from_config('creds', 'AWS_SECRET_ACCESS_KEY', 'secure', 'Please enter AWS_SECRET_ACCESS_KEY: ', profile_name)
    SESSION = get_aws_session(AWS_KEY_ID, AWS_SECRET_ACCESS_KEY, AWS_REGION)
    if SESSION is not None:
        CONFIG.add_aws_profile(AWS_KEY_ID, AWS_SECRET_ACCESS_KEY, AWS_REGION, AWS_OUTPUT, profile_name)
    return SESSION

def update_latest_profile(aws_profile_name):
    BASICS = Config('awstools')
    LATEST = BASICS.get_profile('latest')
    if LATEST is None:
        BASICS.create_profile('latest')
        BASICS.update_config(aws_profile_name, 'role', 'latest')
    else:
        BASICS.update_config(aws_profile_name, 'role', 'latest')

def listToStringWithoutBrackets(list1):
    return str(list1).replace('[','').replace(']','').replace("'", "")

def get_aws_session(aws_key_id, aws_secret_access_key, aws_region):
    CONFIG.unset_aws_profile()
    try:
        SESSION = boto3.Session(
            aws_access_key_id=aws_key_id,
            aws_secret_access_key=aws_secret_access_key,
            region_name=aws_region,
        )
    except:
        return None
    return SESSION

def get_role_session(aws_key_id, aws_secret_access_key, aws_session_token):
    CONFIG.unset_aws_profile()
    try:
        SESSION = boto3.Session(
            aws_access_key_id=aws_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token
        )
    except:
        return None
    return SESSION

def get_env_variables():
    try:
       aws_access_key_id = environ['AWS_ACCESS_KEY_ID']
       aws_secret_access_key = environ['AWS_SECRET_ACCESS_KEY']
       return aws_access_key_id, aws_secret_access_key
    except:
        return None, None

def get_aws_roles(aws_session):
    CLIENT = aws_session.client('iam')
    RESULT = []
    for ROLE in CLIENT.list_roles()['Roles']:
        RESULT.append(ROLE['Arn'])
    return RESULT

def print_role_info(KEY_ID, ACCESS_TOKEN, SESSION_TOKEN):
    EXPORT_CMDS = f"\nexport AWS_ACCESS_KEY_ID={KEY_ID}\nexport AWS_SECRET_ACCESS_KEY={ACCESS_TOKEN}\nexport AWS_SESSION_TOKEN={SESSION_TOKEN}"
    SCRIPT = f"#!/bin/bash\n\n{EXPORT_CMDS}"
    HOME = environ['HOME']
    with open(f"{HOME}/goatrole.sh", 'w') as SCRIPT_FILE:
        SCRIPT_FILE.write(SCRIPT)
        MODE = stat(f"{HOME}/goatrole.sh").st_mode
        MODE |= (MODE & 0o444) >> 2
        chmod(f"{HOME}/goatrole.sh", MODE)
    Log.info("run source ~/goatrole.sh")

def force_cache(aws_profile_name, aws_region_name):
    CONFIG = Config('awstools')
    Log.info('aws profile caching initialized')
    MODULES = ['s3', 'ec2', 'rds']
    for MODULE in MODULES:
        Log.info(f'caching {MODULE} data...')
        run_command(f'goat aws -r {aws_region_name} -p {aws_profile_name} {MODULE} show')

def cache_all_hack(aws_profile_name, aws_region_name):
    CONFIG = Config('awstools')
    Log.info('aws profile caching initialized')
    MODULES = ['s3', 'ec2', 'rds']
    for MODULE in MODULES:
        if MODULE == 's3':
            CACHED = {}
            FOUND = False
            try:
                CACHED.update(CONFIG.get_metadata('cached_buckets', aws_profile_name))
            except:
                pass
            if not CACHED: 
                Log.info(f'caching {MODULE} data...')
                run_command(f'goat aws -r {aws_region_name} -p {aws_profile_name} {MODULE} show')
            elif CACHED:
                for REGION in CACHED:
                    if aws_region_name == REGION:
                        FOUND = True
                        break
                if FOUND is False:
                    Log.info(f'caching {MODULE} data...')
                    run_command(f'goat aws -r {aws_region_name} -p {aws_profile_name} {MODULE} show')
        elif MODULE == 'ec2':
            CACHED = {}
            FOUND = False
            try:
                CACHED.update(CONFIG.get_metadata('cached_instances', aws_profile_name))
            except:
                pass
            if not CACHED:
                Log.info(f'caching {MODULE} data...')
                run_command(f'goat aws -r {aws_region_name} -p {aws_profile_name} {MODULE} show')
            elif CACHED:
                for REGION in CACHED:
                    if aws_region_name == REGION:
                        FOUND = True
                        break
                if FOUND is False:
                    Log.info(f'caching {MODULE} data...')
                    run_command(f'goat aws -r {aws_region_name} -p {aws_profile_name} {MODULE} show')
        elif MODULE == 'rds':
            CACHED = {}
            FOUND = False
            try:
                CACHED.update(CONFIG.PROFILES[aws_profile_name]['metadata']['cached_rds_instances'])
            except:
                pass
            if not CACHED:
                Log.info(f'caching {MODULE} data...')
                run_command(f'goat aws -r {aws_region_name} -p {aws_profile_name} {MODULE} show')
            elif CACHED:
                for REGION in CACHED:
                    if aws_region_name == REGION:
                        FOUND = True
                        break
                if FOUND is False:
                    Log.info(f'caching {MODULE} data...')
                    run_command(f'goat aws -r {aws_region_name} -p {aws_profile_name} {MODULE} show')

def run_command(command):
    PROC = subprocess.run(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)



import sys, os
try:
    import importlib_resources as resources
except:
    from importlib import resources

def get_accounts():

    TARGETS = []
    MY_RESOURCES = resources.files("toolbox")
    DATA = (MY_RESOURCES / "stage_accounts.csv")

    with open(DATA) as f:
        CONTENTS = f.read()
    for line in CONTENTS.split('\n'):
        if line:
            item = line.split(",")[0]
            value = line.split(",")[1]
            if name == item:
                TARGETS.append(line.split(",")[1])
    if TARGETS:
        return TARGETS



import click, json, os
from toolbox.logger import Log
from .s3client import S3client
from configstore.configstore import Config
from toolbox.menumaker import Menu
from toolbox.misc import set_terminal_width
from .iam import get_latest_profile
from pprint import pprint

CONFIG = Config('awstools')
ignore = ['IDP','latest']

@click.group('s3', invoke_without_command=True, help='s3 functions to sync buckets and filesystems', context_settings={'help_option_names':['-h','--help'], 'max_content_width': set_terminal_width()})
@click.option('-m', '--menu', help='use the menu to perform S3 actions', is_flag=True, show_default=True, default=False, required=False)
@click.pass_context
def s3(ctx, menu):
    if menu:
        ctx.obj['MENU'] = True
    else:
        ctx.obj['MENU'] = False

@s3.command(help='create a new bucket', context_settings={'help_option_names':['-h','--help']})
@click.argument('bucket_name', required=True)
@click.pass_context
def create(ctx, bucket_name):
    aws_profile_name = ctx.obj['PROFILE']
    aws_region_name = get_region(ctx, aws_profile_name)
    _create(aws_profile_name, aws_region_name, bucket_name)

def _create(aws_profile_name, aws_region_name, bucket_name):
    S3 = get_S3client(aws_profile_name, aws_region_name)
    RESULT = S3.create_bucket(bucket_name, aws_region_name)
    if RESULT:
        Log.info(f"bucket: {bucket_name} created in region: {aws_region_name} successfully.")
        return RESULT
    else:
        Log.critical("failed to create the bucket")

@s3.command(help='download from s3 to local storage', context_settings={'help_option_names':['-h','--help']})
@click.pass_context
@click.argument('source', required=False, default=None)
@click.option('--destination', help="provide a local destination for download", type=str, required=False, default=os.environ["HOME"], show_default=True)
def download(ctx, source, destination):
    aws_profile_name = ctx.obj['PROFILE']
    aws_region_name = get_region(ctx, aws_profile_name)
    if ctx.obj["MENU"] or source is None:
        if source is None:
            CACHED_BUCKETS = {}
            CACHED_CONTENTS = []
            for PROFILE in CONFIG.PROFILES:
                if PROFILE in ignore:
                    continue
                if PROFILE == aws_profile_name:
                    CACHED_BUCKETS.update(CONFIG.get_metadata_aws('cached_buckets', PROFILE, aws_region_name))
                    CACHED_BUCKETS.pop('last_cache_update', None)
            INPUT = f'Downloads => {aws_profile_name}'
            CHOICE = runMenu(CACHED_BUCKETS, INPUT)
            if CHOICE:
                source = ''.join(CHOICE)
                for PROFILE in CONFIG.PROFILES:
                    if PROFILE in ignore:
                        continue
                    if CONFIG.get_metadata_aws('cached_buckets', PROFILE, aws_region_name) and source in CONFIG.get_metadata_aws('cached_buckets', PROFILE, aws_region_name):
                        aws_profile_name = PROFILE
                        S3 = get_S3client(PROFILE, aws_region_name)
                        CONTENTS, TOTAL, LASTMOD = S3.show_bucket_content(source)
                        if type(CONTENTS) is not bool:
                            for content in CONTENTS:
                                CACHED_CONTENTS.append(content['file'])
                            INPUT = f'Bucket Contents => {aws_profile_name}'
                            CHOICE = runMenu(CACHED_CONTENTS, INPUT)
                            if CHOICE:
                                CHOICE = ''.join(CHOICE)
                                if _download(aws_profile_name, aws_region_name, source, destination, CHOICE):
                                    Log.info(f"downloading {CHOICE} from bucket {source} completed")
                                else:
                                    Log.critical(f"failed to download {CHOICE} from s3 bucket {source} to {destination}")
                            else:
                                Log.critical(f"please select a valid choice to download, exiting.")
                        break
        else:
            CACHED_CONTENTS = []
            for PROFILE in CONFIG.PROFILES:
                if PROFILE in ignore:
                    continue
                if CONFIG.get_metadata_aws('cached_buckets', PROFILE, aws_region_name) and source in CONFIG.get_metadata_aws('cached_buckets', PROFILE, aws_region_name):
                    S3 = get_S3client(PROFILE, aws_region_name)
                    CONTENTS, TOTAL, LASTMOD = S3.show_bucket_content(source)
                    if type(CONTENTS) is not bool:
                        for content in CONTENTS:
                            CACHED_CONTENTS.append(content['file'])
                        INPUT = f'Bucket Contents => {aws_profile_name}'
                        CHOICE = runMenu(CACHED_CONTENTS, INPUT)
                        if CHOICE:
                            CHOICE = ''.join(CHOICE)
                            if _download(aws_profile_name, aws_region_name, source, destination, CHOICE):
                                Log.info(f"downloading {CHOICE} from bucket {source} completed")
                            else:
                                Log.critical(f"failed to download {CHOICE} from s3 bucket {source} to {destination}")
                        else:
                            Log.critical(f"please select a valid choice to download, exiting.")
                    break
    else:
        for aws_profile_name in CONFIG.PROFILES:
            if aws_profile_name in ignore:
                continue
            PROFILE = aws_profile_name
            if CONFIG.get_metadata_aws('cached_buckets', PROFILE, aws_region_name) and source in CONFIG.get_metadata_aws('cached_buckets', PROFILE, aws_region_name):
                S3 = get_S3client(PROFILE, aws_region_name)
                CONTENTS, TOTAL, LASTMOD = S3.show_bucket_content(source)
                if type(CONTENTS) is not bool:
                    for content in CONTENTS:
                        CACHED_CONTENTS.append(content['file'])
                    INPUT = f'Bucket Contents => {aws_profile_name}'
                    CHOICE = runMenu(CACHED_CONTENTS, INPUT)
                    if CHOICE:
                        CHOICE = ''.join(CHOICE)
                        if _download(aws_profile_name, aws_region_name, source, destination, CHOICE):
                            Log.info(f"downloading {CHOICE} from bucket {source} completed")
                        else:
                            Log.critical(f"failed to download {CHOICE} from s3 bucket {source} to {destination}")
                    else:
                        Log.critical(f"please select a valid choice to download, exiting.")
                break

def _download(aws_profile_name, aws_region_name, source, destination, filename):
    S3 = get_S3client(aws_profile_name, aws_region_name)
    try:
        S3.s3_to_local(source, destination, filename)
        return True
    except:
        return False

@s3.command(help='upload from local storage to s3 bucket', context_settings={'help_option_names':['-h','--help']})
@click.option('--source', help="provide a local source for upload", type=str, required=True)
@click.argument('destination', required=False, default=None)
@click.pass_context
def upload(ctx, source, destination):
    aws_profile_name = ctx.obj['PROFILE']
    aws_region_name = get_region(ctx, aws_profile_name)
    CACHED_BUCKETS = {}
    if ctx.obj["MENU"] or destination is None:
        for PROFILE in CONFIG.PROFILES:
            if PROFILE == aws_profile_name:
                CACHED_BUCKETS.update(CONFIG.get_metadata_aws('cached_buckets', PROFILE, aws_region_name))
                CACHED_BUCKETS.pop('last_cache_update', None)
        INPUT = f'Uploads => {aws_profile_name}'
        CHOICE = runMenu(CACHED_BUCKETS, INPUT)
        if CHOICE:
            destination = ''.join(CHOICE)
            for aws_profile_name in CONFIG.PROFILES:
                if aws_profile_name in ignore:
                    continue
                if destination in CONFIG.get_metadata_aws('cached_buckets', aws_profile_name, aws_region_name):
                    if _upload(aws_profile_name, aws_region_name, source, destination):
                        Log.info(f"uploading local files from {source} to s3 bucket {destination} completed")
                    else:
                        Log.critical(f"failed to upload local files from {source} to s3 bucket {destination}")
                    break
    else:
        for PROFILE in CONFIG.PROFILES:
            if PROFILE in ignore:
                continue
            if destination in CONFIG.get_metadata_aws('cached_buckets', PROFILE, aws_region_name):
                aws_profile_name = PROFILE
                if _upload(aws_profile_name, aws_region_name, source, destination):
                    Log.info(f"uploading local files from {source} to s3 bucket {destination} completed")
                else:
                    Log.critical(f"failed to upload local files from {source} to s3 bucket {destination}")
                break

def _upload(aws_profile_name, aws_region_name, source, destination):
    S3 = get_S3client(aws_profile_name, aws_region_name)
    try:
        S3.local_to_s3(destination, source)
        return True
    except:
        return False

@s3.command(help='delete a specified bucket', context_settings={'help_option_names':['-h','--help']})
@click.pass_context
@click.argument('bucket', required=False, default=None)
def delete(ctx, bucket):
    aws_profile_name = ctx.obj['PROFILE']
    aws_region_name = get_region(ctx, aws_profile_name)
    if ctx.obj["MENU"] or bucket is None:
        if bucket is None:
            CACHED_BUCKETS = {}
            for PROFILE in CONFIG.PROFILES:
                if PROFILE in ignore:
                    continue
                if PROFILE == aws_profile_name:
                    CACHED_BUCKETS.update(CONFIG.get_metadata_aws('cached_buckets', PROFILE, aws_region_name))
                    CACHED_BUCKETS.pop('last_cache_update', None)
            INPUT = f'Deletion => {aws_profile_name}'
            CHOICE = runMenu(CACHED_BUCKETS, INPUT)
            if CHOICE:
                bucket = ''.join(CHOICE)
                for PROFILE in CONFIG.PROFILES:
                    if PROFILE in ignore:
                        continue
                    if bucket in CONFIG.get_metadata_aws('cached_buckets', PROFILE, aws_region_name):
                        aws_profile_name = PROFILE
                        _delete(bucket, aws_profile_name, aws_region_name)
                        break
        else:
            for PROFILE in CONFIG.PROFILES:
                if PROFILE in ignore:
                    continue
                if bucket in CONFIG.get_metadata_aws('cached_buckets', PROFILE, aws_region_name):
                    aws_profile_name = PROFILE
                    _delete(bucket, aws_profile_name, aws_region_name)
                    break
    else:
        for PROFILE in CONFIG.PROFILES:
            if PROFILE in ignore:
                continue
            if bucket in CONFIG.get_metadata_aws('cached_buckets', PROFILE, aws_region_name):
                aws_profile_name = PROFILE
                _delete(bucket, aws_profile_name, aws_region_name)
                break

def _delete(bucket, aws_profile_name, aws_region_name):
    S3 = get_S3client(aws_profile_name, aws_region_name)
    RESULT = S3.delete_bucket(bucket, aws_profile_name, aws_region_name)
    if RESULT:
        Log.info(f"bucket: {bucket} deleted in region: {aws_region_name} successfully.")
        return True
    else:
        Log.critical("failed to delete the bucket")

@s3.command(help='manually refresh s3 cached data', context_settings={'help_option_names':['-h','--help']})
@click.pass_context
def refresh(ctx):
    aws_profile_name = ctx.obj['PROFILE']
    aws_region_name = get_region(ctx, aws_profile_name)
    _refresh(aws_profile_name, aws_region_name)

def _refresh(aws_profile_name, aws_region_name):
    try:
        S3 = get_S3client(aws_profile_name, aws_region_name, auto_refresh=False)
        S3.cache_buckets(aws_profile_name)
        return True
    except:
        return False

@s3.command(help='show the data stored in s3 cache', context_settings={'help_option_names':['-h','--help']})
@click.pass_context
@click.argument('bucket', required=False, default=None)
def show(ctx, bucket):
    aws_profile_name = ctx.obj['PROFILE']
    aws_region_name = get_region(ctx, aws_profile_name)
    if ctx.obj["MENU"]:
        if bucket is None:
            CACHED_BUCKETS = {}
            for PROFILE in CONFIG.PROFILES:
                if PROFILE in ignore:
                   continue
                if PROFILE == aws_profile_name:
                    try:
                        CACHED_BUCKETS.update(CONFIG.get_metadata_aws('cached_buckets', PROFILE, aws_region_name))
                        CACHED_BUCKETS.pop('last_cache_update', None)
                    except:
                        Log.critical(f"there are 0 cached buckets for {PROFILE}")
            INPUT = f'Bucket List => {aws_profile_name}'
            CHOICE = runMenu(CACHED_BUCKETS, INPUT)
            if CHOICE:
                CACHED_CONTENTS = []
                for PROFILE in CONFIG.PROFILES:
                    if PROFILE in ignore:
                        continue
                    if CONFIG.get_metadata_aws('cached_buckets', PROFILE, aws_region_name) and ''.join(CHOICE) in CONFIG.get_metadata_aws('cached_buckets', PROFILE, aws_region_name):
                        S3 = get_S3client(PROFILE, aws_region_name)
                        CONTENTS, TOTAL, LASTMOD = S3.show_bucket_content(''.join(CHOICE))
                        if type(CONTENTS) is not bool:
                            for content in CONTENTS:
                                CACHED_CONTENTS.append(content['file'])
                            INPUT = f'Bucket Contents => {aws_profile_name}'
                            CHOICE = runMenu(CACHED_CONTENTS, INPUT)
                        break
        else:
            CACHED_CONTENTS = []
            for PROFILE in CONFIG.PROFILES:
                if PROFILE in ignore:
                    continue
                if bucket in CONFIG.get_metadata_aws('cached_buckets', PROFILE, aws_region_name):
                    S3 = get_S3client(PROFILE, aws_region_name)
                    CONTENTS, TOTAL, LASTMOD = S3.show_bucket_content(bucket)
                    if type(CONTENTS) is not bool:
                        for content in CONTENTS:
                            CACHED_CONTENTS.append(content['file'])
                        INPUT = f'Bucket Contents => {aws_profile_name}'
                        CHOICE = runMenu(CACHED_CONTENTS, INPUT)
                    break
    else:
        if bucket is None:
            _show('buckets', aws_profile_name, aws_region_name)
        else:
            _show('bucket_files', aws_profile_name, aws_region_name, bucket)

def _show(target, aws_profile_name, aws_region_name, bucket=None):
    try:
        if target == 'buckets':
            S3 = get_S3client(aws_profile_name, aws_region_name, cache_only=True)
            S3.show_cache('cached_buckets', aws_profile_name)
        if target == 'bucket_files':
            S3 = get_S3client(aws_profile_name, aws_region_name)
            CONTENTS, TOTAL, LASTMOD = S3.show_bucket_content(bucket)
            if type(CONTENTS) is not bool:
                print(json.dumps(CONTENTS, indent=4, sort_keys=True))
                Log.info(f"bucket: {bucket}, total size: {S3.calculate_object_size(TOTAL)}, and last modified on: {str(LASTMOD)}")
            else:
                Log.critical(f"credentials not available to access s3 bucket {bucket} permission denied, or bucket does not exist")
        return True
    except:
        return False

def get_S3client(aws_profile_name, aws_region='us-east-1', auto_refresh=False, cache_only=False):
    CLIENT = S3client(aws_profile_name, aws_region, False, cache_only)
    if auto_refresh:
        CLIENT.auto_refresh(aws_profile_name)
    return CLIENT

def runMenu(DATA, INPUT):
    COUNT = 0
    FINAL = []
    TITLE = f'S3 Bucket Menu: {INPUT}'
    for data in DATA:
        COUNT = COUNT + 1
        RESULTS = []
        RESULTS.append(data)
        FINAL.append(RESULTS)
    SUBTITLE = f'showing {COUNT} available object(s)'
    JOINER = '\t\t'
    FINAL_MENU = Menu(FINAL, TITLE, JOINER, SUBTITLE)
    CHOICE = FINAL_MENU.display()
    return CHOICE

def get_region(ctx, aws_profile_name):
    AWS_REGION = ctx.obj['REGION']
    if not AWS_REGION:
        S3 = get_S3client(aws_profile_name, 'us-east-1')
        AWS_REGION = S3.get_region_from_profile(aws_profile_name)
    return AWS_REGION



import click, os, json
from toolbox.logger import Log
from .rdsclient import RDSclient
from tabulate import tabulate
from configstore.configstore import Config
from toolbox.misc import set_terminal_width
from .iam import get_latest_profile
from toolbox.menumaker import Menu

CONFIG = Config('awstools')

@click.group('rds', invoke_without_command=True, help='rds to manage the DB instances and related configuration', context_settings={'help_option_names':['-h','--help'], 'max_content_width': set_terminal_width()})
@click.option('-m', '--menu', help='use the menu to perform EC2 actions', is_flag=True, show_default=True, default=False, required=False)
@click.pass_context
def rds(ctx, menu):
    aws_profile_name = ctx.obj['PROFILE']
    if menu:
        ctx.obj['MENU'] = True
    else:
        ctx.obj['MENU'] = False
    pass

@rds.command(help='manually refresh ec2 cached data', context_settings={'help_option_names':['-h','--help']})
@click.pass_context
def refresh(ctx):
    aws_profile_name = ctx.obj['PROFILE']
    aws_region = get_region(ctx, aws_profile_name)
    _refresh('cached_rds_instances', aws_profile_name, aws_region)

def _refresh(cache_type, aws_profile_name, aws_region):
    try:
        if cache_type == 'cached_rds_instances':
            RDS = get_RDSclient(aws_profile_name, aws_region, auto_refresh=False)
            RESULT = RDS.refresh('cached_rds_instances', aws_profile_name)
        return True
    except:
        False
    
@rds.command(help='show the data stored in rds cache', context_settings={'help_option_names':['-h','--help']})
@click.argument('database', required=False)
@click.pass_context
def show(ctx, database):
    aws_profile_name = ctx.obj['PROFILE']
    aws_region = get_region(ctx, aws_profile_name)
    _show(ctx, aws_profile_name, aws_region, database)

def _show(ctx, aws_profile_name, aws_region, database):
    if not database:
        if ctx.obj["MENU"]:
            DATA = []
            RDS = get_RDSclient(aws_profile_name, aws_region, auto_refresh=False, cache_only=True)
            RESPONSE = RDS.get_rds_instances()
            for data in RESPONSE:
                ID = data
                DATA.append(ID)
            INPUT = 'RDS Manager'
            CHOICE = runMenu(DATA, INPUT)
            try:
                CHOICE = ''.join(CHOICE) 
            except:
                Log.critical("please select a DB Identifier to continue...")
            RESPONSE = RDS.describe(CHOICE)
            Log.info(f"describing {CHOICE}:\n" + json.dumps(RESPONSE, indent=2, sort_keys=True, default=str))
        else:
            RDS = get_RDSclient(aws_profile_name, aws_region, auto_refresh=False, cache_only=True)
            RESPONSE = RDS.get_rds_instances()
            show_as_table(RESPONSE, name='RDS TABLE')
    else:
        RDS = get_RDSclient(aws_profile_name, aws_region, auto_refresh=False, cache_only=True)
        RESPONSE = RDS.describe(database)
        Log.info(f"describing {database}:\n" + json.dumps(RESPONSE, indent=2, sort_keys=True, default=str))

def show_as_table(source_data, name=None):
    DATA = []
    for ENTRY in source_data:
        DATA_ENTRY = {}
        if ENTRY != 'last_cache_update':
            DATA_ENTRY['DBInstanceIdentifier'] = ENTRY
            for PROPERTY in source_data[ENTRY]:
                DATA_ENTRY[PROPERTY] = source_data[ENTRY][PROPERTY]
            DATA.append(DATA_ENTRY)
    if DATA != []:
        Log.info(f"{name}\n{tabulate(DATA, headers='keys', tablefmt='rst')}\n")

def get_RDSclient(aws_profile_name, aws_region='us-east-1', auto_refresh=False, cache_only=True):
    CLIENT = RDSclient(aws_profile_name, aws_region, False, cache_only)
    if auto_refresh:
        CLIENT.auto_refresh(aws_profile_name)
    return CLIENT

def runMenu(DATA, INPUT):
    COUNT = 0
    FINAL = []
    TITLE = f'RDS Menu: {INPUT}'
    for data in DATA:
        COUNT = COUNT + 1
        RESULTS = []
        RESULTS.append(data)
        FINAL.append(RESULTS)
    SUBTITLE = f'showing {COUNT} available object(s)'
    JOINER = '\t\t'
    FINAL_MENU = Menu(FINAL, TITLE, JOINER, SUBTITLE)
    CHOICE = FINAL_MENU.display()
    return CHOICE

def get_region(ctx, aws_profile_name):
    AWS_REGION = ctx.obj['REGION']
    if not AWS_REGION:
        RDS = get_RDSclient(aws_profile_name, auto_refresh=False)
        AWS_REGION = RDS.get_region_from_profile(aws_profile_name)
    return AWS_REGION



import datetime, time, os, json, re
from toolbox.logger import Log
from configstore.configstore import Config
from . import iam_nongc
from .aws_config import AWSconfig
from botocore.exceptions import ClientError
from tabulate import tabulate
from pathlib import Path
try:
    import importlib_resources as resources
except:
    from importlib import resources

CONFIGSTORE = Config('awstools')

class S3client():

    def __init__(self, aws_profile_name, aws_region='us-east-1', in_boundary=True, cache_only=False):
        self.CONFIG = AWSconfig()
        self.CONFIGSTORE = Config('awstools')
        self.size_table={0: 'Bs', 1: 'KBs', 2: 'MBs', 3: 'GBs', 4: 'TBs', 5: 'PBs', 6: 'EBs'}
        self.CACHE_UPDATE_INTERVAL = 60*60*24*7 # update cache every week
        self.AWS_PROFILE = aws_profile_name
        self.AWS_REGION = aws_region
        self.INB = in_boundary
        self.CACHE_ONLY = cache_only
        if not self.CACHE_ONLY:
            self.get_session()

    def get_session(self):
        RES = 'non-gc'
        if RES == 'non-gc':
            if self.CONFIG.profile_or_role(self.AWS_PROFILE):
                self.SESSION = iam_nongc._authenticate(self.AWS_PROFILE, self.AWS_REGION)
            else:
                self.SESSION = iam_nongc._assume_role(aws_profile_name=self.AWS_PROFILE)[0]
        return self.SESSION

    def create_bucket(self, bucket_name, aws_region):
        CLIENT = self.SESSION.client('s3')
        try:
            CLIENT.create_bucket(Bucket=bucket_name, CreateBucketConfiguration={'LocationConstraint': aws_region})
            return True
        except ClientError:
            Log.warn('insufficient permissions to create a bucket')
            return False
        except:
            Log.warn('unknown error encountered')
            return False

    def check_policy_status(self, bucket_name):
        CLIENT = self.SESSION.client('s3')
        try:
            RESULT = CLIENT.get_bucket_policy_status(Bucket=bucket_name)
            return RESULT["PolicyStatus"]
        except ClientError:
            Log.warn('insufficient permissions or policy does not exist for this bucket')
            return False
        except:
            Log.warn('unknown error encountered')
            return False

    def get_bucket_policy(self, bucket_name):
        CLIENT = self.SESSION.client('s3')
        try:
            RESULT = CLIENT.get_bucket_policy(Bucket=bucket_name)
            if RESULT:
                return RESULT
            else:
                Log.critical(f"unable to find an attached policy to bucket {bucket_name}")
                return False
        except ClientError: 
            Log.warn("no policy attached to this bucket or insufficient privileges")
            return False
        except:
            Log.warn('unknown error encountered')
            return False

    def show_bucket_content(self, bucket_name, total=0):
        TIMEOUT = time.time() + 90
        ITEMS = []
        LASTMOD = ''
        TOTAL = total
        if self.CACHE_ONLY and hasattr(self, 'SESSION') is False:
            self.get_session()
        CLIENT = self.SESSION.client('s3')
        GET_LAST_MODIFIED = lambda obj: int(obj['LastModified'].strftime('%Y%m%d%H%M%S'))
        try:
            RESPONSE = CLIENT.list_objects(Bucket=bucket_name)
            if 'Contents' in RESPONSE:
                OBJECTS = RESPONSE['Contents']
                FILES = sorted(OBJECTS, key=GET_LAST_MODIFIED, reverse=False)
                for KEY in FILES:
                    if time.time() > TIMEOUT:
                        break
                    FILE = KEY['Key']
                    DATE_MODIFIED = str(KEY["LastModified"])
                    SIZE = KEY['Size']
                    CALC_SIZE = self.calculate_object_size(SIZE)
                    TOTAL = TOTAL + SIZE
                    if 'None' not in FILE:
                        new_dictionary={
                            "file":FILE,
                            "lastModified":DATE_MODIFIED,
                            "size":CALC_SIZE
                        }
                        ITEMS.append(new_dictionary) 
        except:
            return False, False, False
        try:
            LASTMOD = list(ITEMS)[-1]["lastModified"]
        except:
            DATE_LIST = []
            try:
                RESPONSE = CLIENT.list_buckets(Bucket=bucket_name)
                for INDEX in RESPONSE['Buckets']:
                    DATE_LIST.append(INDEX['CreationDate'])
                LASTMOD = list(DATE_LIST)[-1]
            except:
                pass
        return ITEMS, TOTAL, LASTMOD

    def delete_bucket(self, bucket_name, aws_profile_name, aws_region):
        CLIENT = self.SESSION.client('s3')
        try:
            OBJECTS = CLIENT.list_objects_v2(Bucket=bucket_name)
            COUNT = OBJECTS['KeyCount'] 
        except:
            Log.critical(f"unable to find bucket {bucket_name}")
        if COUNT == 0:
            try:
                RESPONSE = CLIENT.delete_bucket(Bucket=bucket_name)
                Log.info(f"bucket {bucket_name} deleted")
            except:
                self.nuke_bucket(bucket_name, CLIENT, aws_region)
        else:
            self.nuke_bucket(bucket_name, CLIENT, aws_region)
        return True

    def get_region_from_profile(self, aws_profile_name):
        AWS_REGION = self.CONFIG.get_from_config('creds', 'region', profile_name=aws_profile_name)
        if AWS_REGION is None: # this is for when the user wants to use a profile which sources another profile for IAM creds
            CREDS_PROFILE = self.CONFIG.get_from_config('creds', 'source_profile', profile_name=aws_profile_name)
            AWS_REGION = self.CONFIG.get_from_config('config', 'region', profile_name=CREDS_PROFILE)
        if AWS_REGION is None:
            Log.critical("Please run goat aws iam assume-role for the target profile before using the s3 module")
        return AWS_REGION

    def nuke_bucket(self, bucket_name, aws_client, aws_region):
        TIMEOUT = time.time() + 600
        SIZE_TABLE = {0: 'Bs', 1: 'KBs', 2: 'MBs', 3: 'GBs', 4: 'TBs', 5: 'PBs', 6: 'EBs'} # appears to be redundant
        # https://wasabi-support.zendesk.com/hc/en-us/articles/360058028992-How-do-I-mass-delete-non-current-versions-inside-a-bucket-
        # include latest verions too so we can nuke the bucket
        PAGINATOR = aws_client.get_paginator('list_object_versions')
        PARAMETERS = {'Bucket': bucket_name}
        DELETE_MARKER_COUNT = 0
        VERSIONED_OBJECT_COUNT = 0
        VERSIONED_OBECT_SIZE = 0
        CURRENT_OBJECT_COUNT = 0
        CURRENT_OBJECT_SIZE = 0
        DELETE_MARKER_LIST = []
        VERSION_LIST = []
        CURRENT_LIST = []
        Log.info("calculating, please wait... this may take a while")
        for ITERATOR in PAGINATOR.paginate(**PARAMETERS):
            if time.time() > TIMEOUT:
                Log.critical(f"unable to delete bucket because time to execute exceeded 10 minutes: {bucket_name}")
            if 'DeleteMarkers' in ITERATOR:
                for DELETE_MARKER in ITERATOR['DeleteMarkers']:
                    DELETE_MARKER_LIST.append({'Key': DELETE_MARKER['Key'], 'VersionId': DELETE_MARKER['VersionId']})
                    DELETE_MARKER_COUNT += 1
                    if 'Versions' in ITERATOR:
                        for VERSION in ITERATOR['Versions']:
                            if VERSION['IsLatest'] is False:
                                VERSIONED_OBJECT_COUNT += 1
                                VERSIONED_OBECT_SIZE += VERSION['Size']
                                VERSION_LIST.append({'Key': VERSION['Key'], 'VersionId': VERSION['VersionId']})
                            elif VERSION['IsLatest'] is True:
                                CURRENT_OBJECT_COUNT += 1
                                CURRENT_OBJECT_SIZE += VERSION['Size']
                                CURRENT_LIST.append({'Key': VERSION['Key'], 'VersionId': VERSION['VersionId']})
            elif 'Versions' in ITERATOR:
                for DELETE_MARKER in ITERATOR['Versions']:
                    DELETE_MARKER_LIST.append({'Key': DELETE_MARKER['Key'], 'VersionId': DELETE_MARKER['VersionId']})
                    DELETE_MARKER_COUNT += 1
                    if 'Versions' in ITERATOR:
                        for VERSION in ITERATOR['Versions']:
                            if VERSION['IsLatest'] is False:
                                VERSIONED_OBJECT_COUNT += 1
                                VERSIONED_OBECT_SIZE += VERSION['Size']
                                VERSION_LIST.append({'Key': VERSION['Key'], 'VersionId': VERSION['VersionId']})
                            elif VERSION['IsLatest'] is True:
                                CURRENT_OBJECT_COUNT += 1
                                CURRENT_OBJECT_SIZE += VERSION['Size']
                                CURRENT_LIST.append({'Key': VERSION['Key'], 'VersionId': VERSION['VersionId']})
        print("-" * 10)
        Log.info(f"total delete markers: {str(DELETE_MARKER_COUNT)}")
        Log.info(f"number of current objects: {str(CURRENT_OBJECT_COUNT)}")
        Log.info(f"current objects size: {self.calculate_object_size(CURRENT_OBJECT_SIZE)}")
        Log.info(f"number of non-current objects: {str(VERSIONED_OBJECT_COUNT)}")
        Log.info(f"non-current objects size: {self.calculate_object_size(VERSIONED_OBECT_SIZE)}")
        Log.info(f"total size of current + non current objects: {self.calculate_object_size(VERSIONED_OBECT_SIZE + CURRENT_OBJECT_SIZE)}")
        print("-" * 10)
        DELETE_FLAG = False
        while not DELETE_FLAG:
            CHOICE = 'y'
            if CHOICE.strip().lower() == 'y':
                DELETE_FLAG = True
                Log.info("starting deletes now...")
                Log.info("removing delete markers 1000 at a time")
                for INDEX in range(0, len(DELETE_MARKER_LIST), 1000):
                    RESPONSE = aws_client.delete_objects(
                        Bucket=bucket_name,
                        Delete={
                            'Objects': DELETE_MARKER_LIST[INDEX:INDEX + 1000],
                            'Quiet': True
                        }
                    )
                Log.info("removing oldest versioned objects 1000 at a time")
                for INDEX in range(0, len(VERSION_LIST), 1000):
                    RESPONSE = aws_client.delete_objects(
                        Bucket=bucket_name,
                        Delete={
                            'Objects': VERSION_LIST[INDEX:INDEX + 1000],
                            'Quiet': True
                        }
                    )
                Log.info("process completed successfully, deleting remaining objects if anything exists")
                try:
                    if self.finish_nuke_bucket(aws_client, bucket_name):
                        try:
                            RESPONSE = aws_client.delete_bucket(Bucket=bucket_name)
                            Log.info(f"bucket: {bucket_name} deleted in region: {aws_region} successfully.")
                        except:
                            Log.critical(f"unable to delete bucket: {bucket_name}")
                    else:
                        Log.critical(f"unable to delete bucket: {bucket_name}")                    
                except:
                    Log.critical(f"unable to delete bucket: {bucket_name}")
        return True

    def finish_nuke_bucket(self, client, bucket_name):
        bucket = client.Bucket(bucket_name)
        try:
            bucket.object_versions.delete()
            return True
        except:
            return False

    def calculate_object_size(self, size):
        count = 0
        while size // 1024 > 0:
            size = size / 1024
            count += 1
        return str(round(size, 2)) + ' ' + self.size_table[count]

    def s3_to_local(self, bucket_name, local_folder, filename):
        CLIENT = self.SESSION.resource('s3')
        BUCKET = CLIENT.Bucket(bucket_name)
        for OBJECT in BUCKET.objects.all():
            s3_path, s3_filename = self.get_s3_path_filename(OBJECT.key)
            base = os.path.basename(filename)
            if filename is not None and base == s3_filename:
                local_folder_path = os.path.join(*[local_folder, s3_path])
                local_fullpath = os.path.join(local_folder_path, s3_filename)
                self.mkdir_p(local_folder_path)
                try:
                    BUCKET.download_file(OBJECT.key, local_fullpath)
                    Log.info(f"finished downloading {OBJECT.key}")
                except:
                    Log.warn(f"problem downloading {OBJECT.key}")
                    pass
            elif filename is None:
                local_folder_path = os.path.join(*[local_folder, s3_path])
                local_fullpath = os.path.join(*[local_folder_path, s3_filename])
                self.mkdir_p(local_folder_path)
                try:
                    BUCKET.download_file(OBJECT.key, local_fullpath)
                    Log.info(f"finished downloading {OBJECT.key}")
                except:
                    Log.warn(f"problem downloading {OBJECT.key}")
                    pass

    def local_to_s3(self, bucket_name, local_path):
        CURRENT_TIME = datetime.datetime.now()
        CLIENT = self.SESSION.client('s3')
        if os.path.isdir(local_path):
            for ROOT, DIRS, FILES in os.walk(local_path):
                for NAME in FILES:
                    FILE_PATH = os.path.join(ROOT, NAME)
                    try:
                        KEY = NAME
                        CLIENT.upload_file(FILE_PATH, bucket_name, KEY)
                        Log.info(f"finished uploading {NAME} to {bucket_name}")
                    except FileExistsError:
                        Log.warn(f"problem uploading {NAME} to {bucket_name}")
                        pass
        else:
            try:
                KEY = local_path
                CLIENT.upload_file(local_path, bucket_name, KEY)
                Log.info(f"finished uploading {local_path} to {bucket_name}")
            except FileExistsError:
                Log.warn(f"problem uploading {local_path} to {bucket_name}")
                pass

    def get_s3_path_filename(self, key):
        key = str(key)
        return key.replace(key.split('/')[-1],""), key.split('/')[-1]

    def mkdir_p(self, path):
        try:
            os.makedirs(path)
        except OSError as exc:  # Python >2.5
            pass

    def auto_refresh(self, aws_profile_name='default'):
        TIME_NOW = datetime.datetime.now().timestamp()
        try:    
            BUCKETS_TS = self.CONFIGSTORE.PROFILES[aws_profile_name]['metadata']['cached_buckets']['last_cache_update']
        except KeyError:
            BUCKETS_TS = 0    
        if TIME_NOW - float(BUCKETS_TS) > self.CACHE_UPDATE_INTERVAL:
            self.cache_buckets(aws_profile_name)
        self.CONFIGSTORE = Config('awstools')

    def refresh(self, aws_profile_name='default'):
        self.cache_buckets(aws_profile_name)
        
    def cache_buckets(self, aws_profile_name='default'):
        TIMESTAMP = str(datetime.datetime.now().timestamp())
        if aws_profile_name not in self.CONFIGSTORE.PROFILES:
            self.CONFIGSTORE.create_profile(aws_profile_name)
        if self.CACHE_ONLY and hasattr(self, 'SESSION') is False:
            self.get_session()
        RESOURCE = self.SESSION.resource('s3')
        Log.info('caching S3 buckets...')
        BUCKETS = [BUCKET.name for BUCKET in RESOURCE.buckets.all()]
        BUCKETS_CACHE = {}
        for BUCKET in BUCKETS:
            BUCKETS_CACHE[BUCKET] = {
                'name': BUCKET
            }
        BUCKETS_CACHE['last_cache_update'] = TIMESTAMP
        DICT = {}
        DICT[self.AWS_REGION] = BUCKETS_CACHE
        self.CONFIGSTORE.update_metadata(DICT, 'cached_buckets', aws_profile_name, False)

    def get_cache(self, type, aws_profile_name, subtype=None):
        if aws_profile_name not in self.CONFIGSTORE.PROFILES:
            return None
        if type not in self.CONFIGSTORE.PROFILES[aws_profile_name]['metadata']:
            return None
        if subtype is None:
            return self.CONFIGSTORE.PROFILES[aws_profile_name]['metadata'][type]
        else:
            if subtype in self.CONFIGSTORE.PROFILES[aws_profile_name]['metadata'][type]:
                return self.CONFIGSTORE.PROFILES[aws_profile_name]['metadata'][type][subtype]
            else:
                return None

    def show_cache(self, type, aws_profile_name):
        self.auto_refresh(aws_profile_name)
        DATA = []
        for ENTRY in self.CONFIGSTORE.PROFILES[aws_profile_name]['metadata'][type][self.AWS_REGION]:
            DATA_ENTRY = {}
            if ENTRY != 'last_cache_update' and ENTRY != 'lastmod' and ENTRY != 'total':
                DATA_ENTRY['Name'] = ENTRY
                DATA.append(DATA_ENTRY)
        Log.info(f"{type}\n{tabulate(DATA, headers='keys', tablefmt='rst')}\n")

    def get_object_tagging(self, bucket_name, key):
        CLIENT = self.SESSION.client('s3')
        RESPONSE = CLIENT.get_object_tagging(Bucket=bucket_name, Key=key)
        return RESPONSE



import click, os, json, re, sys, subprocess
from toolbox.logger import Log
from toolbox.misc import set_terminal_width, get_save_path
from .iam import get_latest_profile
import signal
import atexit
from pathlib import Path

@click.group('extract', invoke_without_command=True, help='extract the command tree for AWS CLI', context_settings={'help_option_names':['-h','--help'], 'max_content_width': set_terminal_width()})
@click.pass_context
def extract(ctx):
    pass

@extract.command(help='manually refresh the AWS CLI command tree', context_settings={'help_option_names':['-h','--help']})
@click.pass_context
def commands(ctx):

    json_file = get_save_path("aws.json")

    # Register cleanup function to handle the SIGINT (Ctrl_C) and SIGTERM signals
    signal.signal(signal.SIGINT, lambda signum, frame: (cleanup(), sys.exit(1)))
    signal.signal(signal.SIGTERM, lambda signum, frame: (cleanup(), sys.exit(1)))

    if not prompt_user_to_continue():
        print("INFO: exiting the script now...")
        exit()

    aws_output = get_aws_help_output()
    cleaned_output = clean_output(aws_output)
    global_options = extract_global_options(cleaned_output)

    start = cleaned_output.find("AVAILABLE SERVICES")
    end = cleaned_output.find("SEE ALSO")

    services = []
    if start != -1 and end != -1:
        service_block = cleaned_output[start:end]
        services = re.findall(r'o (\w+)', service_block, re.MULTILINE)

    aws_data = {
        "aws": {
            "args": [],
            "command": "aws",
            "help": "",
            "options": {},
            "subcommands": {},
        }
    }

    aws_data["aws"]["options"] = global_options
    for service in services:
        if not service.islower() or service == "help":
            continue

        aws_service_output = get_aws_help_output(service)
        if not aws_service_output:
            continue

        commands, _ = extract_elements_for_service(aws_service_output)
        print(f"Exploring {service}")
        if commands:
            aws_data["aws"]["subcommands"][service] = {
                "command": service,
                "help": clean_description(extract_description(aws_service_output)),
                "options": {},
                "subcommands": {}
            }

        for command, command_info in commands.items():
            if not command.islower():
                continue

            print(f"\tExploring {service} {command}")
            aws_command_output = get_aws_help_output(service, command)
            if not aws_command_output:
                continue

            local_options = extract_local_options(aws_command_output)
            if local_options:
                aws_data["aws"]["subcommands"][service]["subcommands"][command] = {
                    "options": local_options,
                    "help": clean_description(extract_description(aws_command_output))
                }

        save_to_json(aws_data, json_file)

def prompt_user_to_continue():
    """Prompts the user to decide if they want to continue the process."""
    print("INFO: this process will take approximately 3-4 hours to complete.")
    choice = input("INFO: do you want to continue? [yes/no]: ").strip().lower()

    if choice == "yes":
        return True
    elif choice == "no":
        return False
    else:
        print("INFO: invalid choice. Please enter 'yes' or 'no'.")
        return prompt_user_to_continue()  # Recursively ask until a valid choice is made.

def save_to_json(data, filename):
    """Save data to a JSON file."""
    with open(filename, 'w') as f:
        json.dump(data, f, indent=4)

def clean_output(text):
    while '\x08' in text:
        idx = text.index('\x08')
        text = text[:idx-1] + text[idx+1:]
    return text

def extract_global_options(aws_output):
    """Extract global options from the AWS CLI help output."""
    lines = aws_output.split("\n")

    options = {}
    current_option = None
    description_lines = []

    for line in lines:
        option_match = re.match(r'^\s*(--[\w-]+)\s+\((\w+)\)', line)
        if option_match:
            if current_option:  # Save the previous option
                options[current_option] = {
                    "name": current_option,
                    "help": " ".join(description_lines).strip()
                }
                description_lines = []

            current_option = option_match.group(1)
        elif current_option:
            description_lines.append(line.strip())

    # Save the last option, if any
    if current_option:
        options[current_option] = {
            "name": current_option,
            "help": " ".join(description_lines).strip()
        }

    return options

def extract_local_options(text):
    options = {}

    # Check if both "OPTIONS" and "GLOBAL OPTIONS" are present, and extract the content between them
    start_index = text.find("OPTIONS")
    end_index = text.find("GLOBAL OPTIONS")

    if start_index == -1 or end_index == -1:
        return options

    local_options_text = text[start_index:end_index]

    lines = local_options_text.split("\n")
    current_option = None
    current_description = []

    for line in lines:
        option_match = re.match(r'^\s*(--[\w-]+)\s+\((\w+)\)', line)
        if option_match:
            if current_option:  # Save the previous option's description
                full_description = ' '.join(current_description).strip()
                first_sentence = extract_first_sentence(full_description)
                options[current_option]["help"] = first_sentence
                current_description = []

            current_option = option_match.group(1)
            option_type = option_match.group(2)
            options[current_option] = {"type": option_type, "help": ""}
        elif current_option:
            current_description.append(line.strip())

    # Save the last option's description, if any
    if current_option and current_description:
        full_description = ' '.join(current_description).strip()
        first_sentence = extract_first_sentence(full_description)
        options[current_option]["help"] = first_sentence

    return options

def extract_elements_for_service(aws_output):
    commands_pattern = re.compile(r'^\s*o\s+([\w-]+)', re.MULTILINE)
    options_pattern = r'(--[\w-]+)\s+\((\w+)\)'

    commands = {}
    options = {}

    current_service = None
    current_command = None
    current_option = None
    option_description = []

    for line in aws_output.split("\n"):
        command_match = re.search(commands_pattern, line)
        if command_match:
            current_service = command_match.group(1)
            commands[current_service] = {"command": current_service, "options": {}, "subcommands": {}}
            current_command = None
            continue

        subcommand_match = re.search(r'^\s*o\s+([a-zA-Z0-9-]+)', line)
        if subcommand_match:
            current_command = subcommand_match.group(1)
            commands[current_service]["subcommands"][current_command] = {"command": current_command, "options": {}}
            continue

        option_match = re.search(options_pattern, line)
        if option_match:
            # Store the previous option's description, if any
            if current_option:
                options[current_option]['help'] = ' '.join(option_description).strip()
                option_description = []

            current_option = option_match.group(1)
            option_type = option_match.group(2)
            options[current_option] = {"name": current_option, "type": option_type, "help": ""}
            continue

        if current_option:
            option_description.append(line.strip())

    # Save the last option's description, if any
    if current_option and option_description:
        options[current_option]['help'] = ' '.join(option_description).strip()

    return commands, options

def extract_description(aws_output):
    description_pattern = re.compile(r'\s*DESCRIPTION\s*([\s\S]*?)\n\n', re.MULTILINE)
    match = re.search(description_pattern, aws_output)
    if match:
        description = match.group(1).strip()
        return description
    return ""

def get_aws_help_output(service="", command=""):
    try:
        if command:
            aws_help_output = subprocess.check_output(["aws", service, command, "help"], text=True, stderr=subprocess.STDOUT)
        elif service:
            aws_help_output = subprocess.check_output(["aws", service, "help"], text=True, stderr=subprocess.STDOUT)
        else:
            aws_help_output = subprocess.check_output(["aws", "help"], text=True, stderr=subprocess.STDOUT)
        return clean_output(aws_help_output)
    except subprocess.CalledProcessError as e:
        print(f"An error occurred while fetching AWS CLI help: {str(e)}")
        return ""

def cleanup(filename="aws.json"):
    save_path = get_save_path(filename)
    if os.path.isfile(filename):
        os.remove(filename)

def clean_description(desc):
    # Normalize whitespace
    desc = ' '.join(desc.split())

    # Extract the first sentence
    match = re.match(r'^(.*?[.!?])', desc)
    if match:
        return match.group(1)
    return desc

def extract_first_sentence(description):
    match = re.match(r"^(.*?[.!?])", description)
    if match:
        return match.group(1).strip()
    return description.strip()
